@@ -51,6 +51,8 @@ from datetime import timedelta
 import time
 from pprint import pprint
 from operator import itemgetter
+import logging
+from collections import defaultdict
 # python 2.6 support
 try:
     from collections import OrderedDict
@@ -69,6 +71,9 @@ try:
 except:
     from exoline import __version__
 
+logging.basicConfig(stream=sys.stderr)
+logging.getLogger("pyonep.onep").setLevel(logging.ERROR)
+
 DEFAULT_HOST = 'm2.exosite.com'
 DEFAULT_PORT = '80'
 DEFAULT_PORT_HTTPS = '443'
@@ -81,8 +86,8 @@ cmd_doc = {
 Options:
     --follow                 continue reading (ignores --end)
     --limit=<limit>          number of data points to read [default: 1]
-    --start=<time>           start time (see details below)
-    --end=<time>             end time
+    --start=<time>
+    --end=<time>             start and end times (see details below)
     --selection=all|autowindow|givenwindow  downsample method [default: all]
     --format=csv|raw         output format [default: csv]
     --timeformat=unix|human  unix timestamp or human-readable? [default: human]
@@ -120,22 +125,32 @@ Options:
     {{ helpoption }}
 
 Details:
-    Pass - and a json description object on stdin for maximum control.
+    Pass - and a json description object on stdin, or leave it off to use defaults.
     Description is documented here:
     http://developers.exosite.com/display/OP/Remote+Procedure+Call+API#RemoteProcedureCallAPI-create
 
     If - is not present, creates a resource with common defaults.''',
     'listing':
         '''List a client's children based on their type.\n\nUsage:
-    exo [options] listing <cik> (--type=client|dataport|datarule|dispatch) ... [--plain] [--pretty]''',
+    exo [options] listing <cik> (--type=client|dataport|datarule|dispatch) ...
+
+Options:
+    --plain   show only the child RIDs
+    --pretty  pretty print output''',
     'info':
-        '''Get info for a resource in json format.\n\nUsage:
+        '''Get metadata for a resource in json format.\n\nUsage:
     exo [options] info <cik> [<rid>]
 
 Options:
-    --cikonly    print CIK by itself
-    --pretty     pretty print output
-    --recursive  embed info for any children recursively''',
+    --cikonly      print CIK by itself
+    --pretty       pretty print output
+    --recursive    embed info for any children recursively
+    --include=<key list>
+    --exclude=<key list>
+                   comma separated list of info keys to include and exclude.
+                   Available keys are aliases, basic, counts, description,
+                   key, shares, storage, subscribers, tags, usage. If omitted,
+                   all available keys are returned.''',
     'update':
         '''Update a resource from a json description passed on stdin.\n\nUsage:
     exo [options] update <cik> [<rid>] -
@@ -188,13 +203,18 @@ Options:
     {{ startend }}''',
     'tree':
         '''Display a resource's descendants.\n\nUsage:
-    exo tree [--verbose] [--hide-keys] <cik>''',
-    'script': '''Upload a Lua script\n\nUsage:
+    exo [options] tree [--verbose] [--hide-keys] <cik>
+
+    --counts       show item counts (from info.storage.counts)
+    --level=<num>  depth to traverse, omit or -1 for no limit [default: -1]''',
+    'script': '''Upload a client's a Lua script\n\nUsage:
     exo [options] script <script-file> <cik> ...
 
 Options:
-    --name=<name>  script name, if different from script filename.
-    {{ helpoption }}''',
+    --name=<name>  script name, if different from script filename. The name
+                   is used to identify the script, too.
+    --recursive    operate on client and any children
+    --create       create the script if it doesn't already exist''',
     'intervals': '''Show distribution of intervals between points.\n\nUsage:
     exo [options] intervals <cik> [<rid>] --days=<days>
 
@@ -231,13 +251,13 @@ doc_replace = {
 
     If time part is omitted, it assumes 00:00:00.
     To report through the present time, omit --end or pass --end=now''',
-    '{{ helpoption }}': '''    -h --help            Show this screen.''',
+    '{{ helpoption }}': '''-h --help  Show this screen.''',
 }
 
 for k in cmd_doc:
     # helpoption is appended to any commands that don't already have it
     if '{{ helpoption }}' not in cmd_doc[k]:
-        cmd_doc[k] += '\n\nOptions:\n    {{ helpoption }}'
+        cmd_doc[k] += '\n\nOptions:\n{{ helpoption }}'
     for r in doc_replace:
         cmd_doc[k] = cmd_doc[k].replace(r, doc_replace[r])
 
@@ -274,8 +294,9 @@ class ExoRPC():
             raise RPCException(msg)
 
     def _raise_for_response_record(self, isok, response):
-        '''Undocumented RPC behavior-- if record timestamps are invalid, isok is True
-           but response is an array of timestamps and error messages.'''
+        '''Undocumented RPC behavior-- if record timestamps are invalid, isok
+           is True but response is an array of timestamps and error
+           messages.'''
         self._raise_for_response(isok, response)
         if type(response) is list:
             # TODO: does this always indicate an error condition?
@@ -529,7 +550,7 @@ class ExoRPC():
         self._raise_for_response(isok, response)
         return response
 
-    def _listing_with_info(self, cik, types):
+    def _listing_with_info(self, cik, types, options={}):
         '''Return a list of dicts for each type containing mappings between rids and
         info for that RID. E.g.:
         [{'<rid0>':<info0>, '<rid1>':<info1>}, {'<rid2>':<info2>}, [], {'<rid3>': <info3>}]'''
@@ -544,7 +565,7 @@ class ExoRPC():
         # request info for each rid
         # (rids is a flattened version of listing)
         rids = list(itertools.chain(*listing))
-        responses = self._exomult(cik, [['info', rid] for rid in rids])
+        responses = self._exomult(cik, [['info', rid, options] for rid in rids])
 
         # From the return values make a list of dicts like this:
         # [{'<rid0>':<info0>, '<rid1>':<info1>}, {'<rid2>':<info2>}, [], {'<rid3>': <info3>}]
@@ -560,10 +581,16 @@ class ExoRPC():
 
         return listing_with_info
 
-    def info(self, cik, rid={'alias': ''}, options={}, cikonly=False, recursive=False):
+    def info(self,
+             cik,
+             rid={'alias': ''},
+             #options={"usage": False, "counts": False},
+             options={},
+             cikonly=False,
+             recursive=False):
         if recursive:
             rid = None if type(rid) is dict else rid
-            response = self._infotree(cik, rid=rid)
+            response = self._infotree(cik, rid=rid, options=options)
         else:
             isok, response = self.exo.info(cik, rid, options)
             self._raise_for_response(isok, response)
@@ -665,15 +692,22 @@ class ExoRPC():
             u'' if len(opt) == 0 else u'({0})'.format(u', '.join(
                 [u'{0}: {1}'.format(k, v) for k, v in opt.iteritems()]))))
 
-    def tree(self, cik, aliases=None, cli_args={}, spacer=u''):
+    def tree(self, cik, aliases=None, cli_args={}, spacer=u'', level=0, info_options={}):
         '''Print a tree of entities in OneP'''
+        max_level = int(cli_args['--level'])
         # print root node
         isroot = len(spacer) == 0
         if isroot:
+            # usage and counts are slow, so omit them if we don't need them
+            exclude = ['usage']
+            if not cli_args['--counts']:
+                exclude.append('counts')
+                exclude.append('storage')
+            info_options = self.make_info_options(exclude=exclude)
             # todo: combine these (maybe also the _listing_with_info function
             # below?)
             rid = self.lookup(cik, "")
-            info = self.info(cik)
+            info = self.info(cik, options=info_options)
             # info doesn't contain key
             info['key'] = cik
             aliases = info['aliases']
@@ -685,10 +719,13 @@ class ExoRPC():
                              cli_args,
                              spacer,
                              islast=True)
+            if max_level == 0:
+                return
+            level += 1
 
         types = ['dataport', 'datarule', 'dispatch', 'client']
         try:
-            listing = self._listing_with_info(cik, types=types)
+            listing = self._listing_with_info(cik, types=types, options=info_options)
             # listing(): [['<rid0>', '<rid1>'], ['<rid2>'], [], ['<rid3>']]
             # _listing_with_info(): [{'<rid0>':<info0>, '<rid1>':<info1>},
             #                       {'<rid2>':<info2>}, [], {'<rid3>': <info3>}]
@@ -713,7 +750,8 @@ class ExoRPC():
                     if t == 'client':
                         next_cik = info['key']
                         self._print_node(rid, info, aliases, cli_args, own_spacer, islast)
-                        self.tree(next_cik, info['aliases'], cli_args, child_spacer)
+                        if max_level == -1 or level < max_level:
+                            self.tree(next_cik, info['aliases'], cli_args, child_spacer, level=level + 1, info_options=info_options)
                     else:
                         self._print_node(rid, info, aliases, cli_args, own_spacer, islast)
 
@@ -774,7 +812,22 @@ class ExoRPC():
             else:
                 raise ExoException("Error updating datarule.")
 
-    def upload(self, ciks, filename, name=None):
+    def cik_recursive(self, cik, fn):
+        '''Run fn on cik and all its client children'''
+        fn(cik)
+        lwi = self._listing_with_info(cik,
+                                      ['client'],
+                                      {'key': True})
+        # [{'<rid0>':<info0>, '<rid1>':<info1>}]
+        for rid in lwi[0]:
+            self.cik_recursive(lwi[0][rid]['key'], fn)
+
+    def upload_script(self,
+                      ciks,
+                      filename,
+                      name=None,
+                      recursive=False,
+                      create=False):
         try:
             f = open(filename)
         except IOError:
@@ -786,15 +839,25 @@ class ExoRPC():
                     # if no name is specified, use the file name as a name
                     name = os.path.basename(filename)
                 for cik in ciks:
-                    rid = self._lookup_rid_by_name(cik, name)
-                    self._upload_script(cik, name, text, rid)
+                    def up(cik):
+                        rid = self._lookup_rid_by_name(cik, name)
+                        print("create: {0}".format(create))
+                        if rid is not None or create:
+                            self._upload_script(cik, name, text, rid=rid)
+                        else:
+                            print("Skipping CIK: {0} -- script named {1} not found".format(cik, name))
+
+                    if recursive:
+                        self.cik_recursive(cik, up)
+                    else:
+                        up(cik)
 
     def lookup_rid(self, cik, cik_to_find):
         isok, listing = self.exo.listing(cik, types=['client'])
         self._raise_for_response(isok, listing)
 
         for rid in listing[0]:
-            self.exo.info(cik, rid, defer=True)
+            self.exo.info(cik, rid, {'key': True}, defer=True)
 
         if self.exo.has_deferred(cik):
             responses = self.exo.send_deferred(cik)
@@ -823,21 +886,18 @@ class ExoRPC():
 
 
     def _create_from_infotree(self, parentcik, infotree):
-        if 'basic' not in infotree:
-            # we're initially passed infotree with a single key
-            infotree = infotree[infotree.keys()[0]]
-        typ = infotree['basic']['type']
-        rid = self.create(parentcik, typ, infotree['description'])
+        '''Create a copy of infotree under parentcik'''
+        typ = infotree['info']['basic']['type']
+        rid = self.create(parentcik, typ, infotree['info']['description'])
         if typ == 'client':
             # look up new CIK
             cik = self.info(parentcik, rid)['key']
-            children = infotree['children']
+            children = infotree['info']['children']
             aliases_to_create = {}
-            for childrid in children:
-                childinfotree = children[childrid]
-                newrid, _ = self._create_from_infotree(cik, childinfotree)
-                if childrid in infotree['aliases']:
-                    aliases_to_create[newrid] = infotree['aliases'][childrid]
+            for child in children:
+                newrid, _ = self._create_from_infotree(cik, child)
+                if child['rid'] in infotree['info']['aliases']:
+                    aliases_to_create[newrid] = infotree['info']['aliases'][child['rid']]
 
             # add aliases in one request
             self._exomult(
@@ -849,13 +909,40 @@ class ExoRPC():
         else:
             return rid, None
 
+    def _counttypes(self, infotree, counts=defaultdict(int)):
+        '''Return a dictionary with the count of each type of resource in the
+        tree. For example, {'client': 2, 'dataport': 1, 'dispatch':1}'''
+        info = infotree['info']
+        counts[info['basic']['type']] += 1
+        if 'children' in info:
+            for child in info['children']:
+                counts = self._counttypes(child, counts=counts)
+        return counts
+
     def copy(self, cik, destcik, infotree=None):
         '''Make a copy of cik and its non-client children to destcik and
         return the cik of the copy.'''
 
         # read in the whole client to copy at once
         if infotree is None:
-            infotree = self._infotree(cik)
+            infotree = self._infotree(cik, options={})
+
+        # check counts
+        counts = self._counttypes(infotree)
+        destinfo = self.info(destcik)
+
+        noroom = ''
+        for typ in counts:
+            destlimit = destinfo['description']['limits'][typ]
+            destcount = destinfo['counts'][typ]
+            needs = counts[typ]
+            # TODO: need a way to check if limit is set to 'inherit'
+            if type(destlimit) is int and destlimit - destcount < needs:
+                noroom = noroom + 'Thing to copy has {0} {1}{4}, parent has limit of {3} (and is using {2}).\n'.format(
+                    needs, typ, destcount, destlimit, 's' if needs > 1 else '')
+
+        if len(noroom) > 0:
+            raise ExoException('Copy would violate parent limits:\n{0}'.format(noroom))
 
         cprid, cpcik = self._create_from_infotree(destcik, infotree)
 
@@ -894,25 +981,26 @@ class ExoRPC():
 
         return list(differ.compare(s1, s2))
 
-    def _infotree(self, cik, rid=None, nodefn=lambda rid, info: rid):
+    def _infotree(self, cik, rid=None, nodeidfn=lambda rid, info: rid, options={}):
         '''Get all info for a cik and its children in a nested dict.
-           The basic unit is 'rid: xyz': <info-with-children>, where <info-with-children>
-           is just the info object for that node with the addition of 'children' key,
-           which is a dict containing more nodes. Here's an example return value:
+        The basic unit is {'rid': '<rid>', 'info': <info-with-children>},
+        where <info-with-children> is just the info object for that node
+        with the addition of 'children' key, which is a dict containing
+        more nodes. Here's an example return value:
 
-           {'<rid 0>': {'description': ...,
+           {'rid': '<rid 0>', 'info': {'description': ...,
                         'basic': ....,
                         ...
-                        'children: {'<rid 1>': {'description': ...,
+                        'children: [{'rid': '<rid 1>', 'info': {'description': ...,
                                                 'basic': ...
-                                                'children': {'<rid 2>': {'description': ...,
+                                                'children': [{'rid': '<rid 2>', 'info': {'description': ...,
                                                                          'basic': ...,
-                                                                         'children: {} } } },
-                                    '<rid 3>': {'description': ...,
+                                                                         'children: [] } } },
+                                    {'rid': '<rid 3>', 'info': {'description': ...,
                                                 'basic': ...
-                                                'children': {} } } } }
+                                                'children': {} } }] } }
 
-           As it's building this nested dict, it calls nodefn with the rid and info
+           As it's building this nested dict, it calls nodeidfn with the rid and info
            (w/o children) for each node.
         '''
         types = ['client', 'dataport', 'datarule', 'dispatch']
@@ -923,7 +1011,7 @@ class ExoRPC():
         if norid:
             rid = self._exomult(cik, [['lookup', 'aliased', '']])[0]
 
-        info = self._exomult(cik, [['info', rid]])[0]
+        info = self._exomult(cik, [['info', rid, options]])[0]
 
         if norid or info['basic']['type'] == 'client':
             if not norid:
@@ -932,19 +1020,22 @@ class ExoRPC():
             listing = self._exomult(cik,
                                     [['listing', types]])[0]
 
-        myid = nodefn(rid, info)
+        myid = nodeidfn(rid, info)
 
-        info['children'] = {}
+        info['children'] = []
         for typ, ridlist in zip(types, listing):
             for childrid in ridlist:
-                tr = self._infotree(cik, childrid, nodefn=nodefn)
-                info['children'].update(tr)
+                tr = self._infotree(cik, childrid, nodeidfn=nodeidfn, options=options)
+                info['children'].append(tr)
+        info['children'].sort(key=itemgetter('rid'))
 
-        return {myid : info}
+        return {'rid': myid, 'info': info}
 
     def _difffilter(self, difflines):
         d = difflines
 
+        # TODO: fix this for new infotree format
+
         # replace differing rid children lines with a single <<rid>>
         ridline = '^[+-](.*").*\.[a-f0-9]{40}(".*)\n'
         d = re.sub(ridline * 2, r' \1<<RID>>\2\n', d, flags=re.MULTILINE)
@@ -997,8 +1088,8 @@ class ExoRPC():
                 # prepend the name so that node names tend to sort (and so
                 # compare well)
                 return info['description']['name'] + '.' + rid
-            info1 = self._infotree(cik1, nodefn=name_prepend)
-            info2 = self._infotree(cik2, nodefn=name_prepend)
+            info1 = self._infotree(cik1, nodeidfn=name_prepend, options={})
+            info2 = self._infotree(cik2, nodeidfn=name_prepend, options={})
 
         if info1 == info2:
             return None
@@ -1017,6 +1108,38 @@ class ExoRPC():
 
             return differences
 
+    def make_info_options(self, include=[], exclude=[]):
+        '''Create options for the info command based on included
+        and excluded keys.'''
+        options = {}
+        # TODO: this is a workaround. The RPC API returns empty list if any
+        # keys are set to false. So, the workaround is to include all keys
+        # except for the excluded ones. This has the undesirable
+        # side-effect of producing "<key>": null in the results, so it would be
+        # better for this to be done in the API.
+        #
+        #for key in exclude:
+        #    options[key] = False
+
+        if len(exclude) > 0:
+            options.update(dict([(k, True) for k in ['aliases',
+                                                        'basic',
+                                                        'counts',
+                                                        'description',
+                                                        'key',
+                                                        'shares',
+                                                        'storage',
+                                                        'subscribers',
+                                                        'tags',
+                                                        'usage']
+                                    if k not in exclude]))
+        else:
+            for key in include:
+                options[key] = True
+
+        return options
+
+
 def parse_ts(s):
     return int(time.mktime(parser.parse(s).timetuple()))
 
@@ -1177,7 +1300,9 @@ def show_intervals(er, cik, rid, start, end, limit, numstd=None):
     sys.stdout.write(' ' * (num_bins - len(min_label) - len(max_label)))
     sys.stdout.write(max_label + '\n')
 
+
 def read_cmd(er, cik, rids, args):
+    '''Read command'''
     limit = args['--limit']
     limit = 1 if limit is None else int(limit)
 
@@ -1203,18 +1328,21 @@ def read_cmd(er, cik, rids, args):
 
     sleep_seconds = 2
     if args['--follow']:
+        if len(rids) > 1:
+            raise ExoException('--follow does not support reading from multiple rids')
         results = []
         while len(results) == 0:
             results = er.readmult(cik,
                                   rids,
                                   limit=1,
                                   sort='desc')
+            # --follow doesn't want the result to be an iterator
+            results = list(results)
             if len(results) > 0:
-                last_t, last_v = results[-1]
-                printline(last_t, last_v)
+                for last_t, last_v in results:
+                    printline(last_t, last_v)
             else:
                 time.sleep(sleep_seconds)
-
         while True:
             results = er.readmult(cik,
                                   rids,
@@ -1222,6 +1350,7 @@ def read_cmd(er, cik, rids, args):
                                   # read time. Could also be now - last_t?
                                   limit=sleep_seconds * 3,
                                   starttime=last_t + 1)
+            results = list(results)
 
             for t, v in results:
                 printline(t, v)
@@ -1244,9 +1373,15 @@ def read_cmd(er, cik, rids, args):
                                 chunkhours=chunkhours):
             printline(t, v)
 
+
 def plain_print(arg):
     print(arg)
 
+
+def pretty_print(arg):
+    print(json.dumps(arg, sort_keys=True, indent=4, separators=(',', ': ')))
+
+
 def handle_args(cmd, args):
     er = ExoRPC(host=args['--host'], port=args['--port'], https=args['--https'], httptimeout=args["--httptimeout"])
     cik = args['<cik>']
@@ -1271,7 +1406,7 @@ def handle_args(cmd, args):
                 rids.append(rid_or_alias(args['<rid>']))
 
     if args.get('--pretty', False):
-        pr = pprint
+        pr = pretty_print
     else:
         pr = plain_print
 
@@ -1391,12 +1526,24 @@ def handle_args(cmd, args):
         listing = er.listing(cik, types)
         if args['--plain']:
             for l in listing:
-                for cik in listing[0]:
-                    print(cik)
+                for rid in listing[0]:
+                    print(rid)
         else:
             pr(listing)
     elif cmd == 'info':
-        info = er.info(cik, rids[0], cikonly=args['--cikonly'], recursive=args['--recursive'])
+        include = args['--include']
+        include = [] if include is None else [key.strip()
+                                              for key in include.split(',')]
+        exclude = args['--exclude']
+        exclude = [] if exclude is None else [key.strip()
+                                              for key in exclude.split(',')]
+
+        options = er.make_info_options(include, exclude)
+        info = er.info(cik,
+                       rids[0],
+                       options=options,
+                       cikonly=args['--cikonly'],
+                       recursive=args['--recursive'])
         if args['--pretty']:
             pr(info)
         else:
@@ -1421,7 +1568,10 @@ def handle_args(cmd, args):
         er.tree(cik, cli_args=args)
     elif cmd == 'script':
         # cik is a list of ciks
-        er.upload(cik, args['<script-file>'], args['--name'])
+        er.upload_script(cik, args['<script-file>'],
+                         name=args['--name'],
+                         recursive=args['--recursive'],
+                         create=args['--create'])
     elif cmd == 'intervals':
         days = int(args['--days'])
         end = time.mktime(datetime.now().timetuple())
@@ -1437,6 +1587,9 @@ def handle_args(cmd, args):
         else:
             pr('cik: ' + newcik)
     elif cmd == 'diff':
+        if sys.version_info < (2, 7):
+            raise ExoException('diff command requires Python 2.7 or above')
+
         diffs = er.diff(cik,
                         args['<cik2>'],
                         full=args['--full'],
