@@ -8,6 +8,7 @@ import sys
 import json
 import re
 import time
+from datetime import datetime
 import StringIO
 import logging
 from unittest import TestCase
@@ -20,7 +21,6 @@ except:
     sys.stderr.write(
         "Copy testconfig.py.template to testconfig.py and set portalcik.")
 
-
 class CmdResult():
     def __init__(self, exitcode, stdout, stderr):
         self.exitcode = exitcode
@@ -33,17 +33,18 @@ logging.getLogger("_cmd").setLevel(logging.DEBUG)
 logging.getLogger("pyonep.onep").setLevel(logging.ERROR)
 log = logging.getLogger("_cmd")
 
+def abbrev(s, length=1000):
+    if len(s) > length:
+        s = s[:length/2] + '\n...\n' + s[-length/2:]
+    return s
 
-def _cmd(argv, stdin):
+def _cmd(argv, stdin, block=True):
     '''Runs an exoline command, translating stdin from
     string and stdout to string. Returns a CmdResult.'''
     if True:
         log.debug(' '.join([str(a) for a in argv]))
         if stdin is not None:
-            display = stdin
-            if len(display) > 2000:
-                display = display[:1000] + '\n...\n' + display[-1000:]
-            log.debug('    stdin: ' + display)
+            log.debug('    stdin: ' + abbrev(stdin))
     if type(stdin) is str:
         sio = StringIO.StringIO()
         sio.write(stdin)
@@ -55,13 +56,10 @@ def _cmd(argv, stdin):
     # unicode causes problems in docopt
     argv = [str(a) for a in argv]
     exitcode = exo.cmd(argv=argv, stdin=stdin, stdout=stdout, stderr=stderr)
-
     stdout.seek(0)
     stdout = stdout.read().strip()  # strip to get rid of leading newline
     stderr.seek(0)
     stderr = stderr.read().strip()
-    if exitcode != 0:
-        log.debug("Exit code was {0}".format(exitcode))
     return CmdResult(exitcode, stdout, stderr)
 
 
@@ -85,6 +83,12 @@ class Resource():
                                       "duration": "infinity"}
             self.desc['public'] = False
 
+    def __str__(self):
+        return 'Resource (parent {0}, type {1}, desc {2})'.format(self.parentcik, self.type, self.desc)
+
+    def __repr__(self):
+        return str(self)
+
     def created(self, rid, info):
         self.rid = rid
         self.info = info
@@ -95,12 +99,46 @@ class Resource():
 
 class TestRPC(TestCase):
     RE_RID = '[0-9a-f]{40}'
+
+    def _logall(self, r):
+        self.l('stdout: {0}\nstderr: {1}'.format(abbrev(r.stdout), abbrev(r.stderr)))
+
+    def _stdre(self, r, msg, search, match, stderr=False):
+        std, label = (r.stderr, "stderr") if stderr else (r.stdout, "stdout")
+        if search is not None:
+            self.assertTrue(re.search(search, std, flags=re.MULTILINE) is not None,
+                msg + ' - failed to find {0}\n'.format(search) + label + ':' + std + '\nsearch expression: ' + search)
+        if match is not None:
+            self.assertTrue(re.match(match, std, flags=re.MULTILINE) is not None,
+                msg + ' - failed to match {0}\n'.format(match) + label + ':' + std + '\nmatch expression: ' + match)
+
+    def notok(self, response, msg='', search=None, match=None):
+        if response.exitcode == 0:
+            self._logall(response)
+        self.assertNotEqual(response.exitcode, 0, msg + ' (exit code should not be 0)')
+        self._stdre(response, msg, search=search, match=match, stderr=True)
+
+    def ok(self, response, msg='', search=None, match=None):
+        if response.exitcode != 0:
+            self._logall(response)
+        self.assertEqual(response.exitcode, 0, msg + ' (exit code should be 0)')
+        self._stdre(response, msg, search=search, match=match, stderr=False)
+
     def _rid(self, s):
         '''Parse rid from s, raising an exception if it doesn't validate.'''
         m = re.match("^({0}).*".format(self.RE_RID), s)
         self.assertFalse(m is None, "rid: {0}".format(s))
         return str(m.groups()[0])
 
+    def _ridcik(self, s):
+        '''Parse rid and cik from output of create command, and raise
+        exception if it doesn't validate.'''
+        m = re.match("^rid: ({0})\ncik: ({0})$".format(self.RE_RID),
+                     s,
+                     re.MULTILINE)
+        self.l(s)
+        return [str(g) for g in m.groups()]
+
     def _createMultiple(self, cik, resList):
         # use pyonep directly
         pyonep = exo.ExoRPC().exo
@@ -109,31 +147,29 @@ class TestRPC(TestCase):
 
         rids = []
         # create resources
-        if pyonep.has_deferred(cik):
-            responses = pyonep.send_deferred(cik)
-            for i, trio in enumerate(responses):
-                call, isok, response = trio
-                if not isok:
-                    raise Exception("_createMultiple failed create()")
-                # response is an rid
-                rid = response
-                rids.append(rid)
-                pyonep.info(cik, rid, defer=True)
+        responses = pyonep.send_deferred(cik)
+        for i, trio in enumerate(responses):
+            call, isok, response = trio
+            self.l('{0}'.format(resList))
+            self.assertTrue(isok, "create should succeed")
+            # response is an rid
+            rid = response
+            rids.append(rid)
+            pyonep.info(cik, rid, defer=True)
 
         # get info
-        if pyonep.has_deferred(cik):
-            responses = pyonep.send_deferred(cik)
-            for i, trio in enumerate(responses):
-                call, isok, response = trio
-                if not isok:
-                    raise Exception("_createMultiple failed info()")
-                # response is info
-                info = response
-                resList[i].created(rids[i], info)
-                res = resList[i]
-                if res.alias is not None:
-                    pyonep.map(cik, resList[i].rid, res.alias, defer=True)
-                self.l("Created {0}, rid: {1}".format(res.type, res.rid))
+        responses = pyonep.send_deferred(cik)
+        for i, trio in enumerate(responses):
+            call, isok, response = trio
+            if not isok:
+                raise Exception("_createMultiple failed info()")
+            # response is info
+            info = response
+            resList[i].created(rids[i], info)
+            res = resList[i]
+            if res.alias is not None:
+                pyonep.map(cik, resList[i].rid, res.alias, defer=True)
+            self.l("Created {0}, rid: {1}".format(res.type, res.rid))
 
         # map to aliases
         if pyonep.has_deferred(cik):
@@ -143,6 +179,8 @@ class TestRPC(TestCase):
                 if not isok:
                     raise Exception("_createMultiple failed map()")
 
+        return rids
+
     def _create(self, res):
         '''Creates a resource at the command line.'''
         alias = [] if res.alias is None else [res.alias]
@@ -156,7 +194,7 @@ class TestRPC(TestCase):
         self.assertEqual(r.exitcode, 0, 'create succeeds')
 
         rid = re.match('rid: ({0})'.format(self.RE_RID), r.stdout).groups()[0]
-        ri = rpc('info', res.parentcik, rid)
+        ri = rpc('info', res.parentcik, rid, '--exclude=count,usage')
         info = json.loads(ri.stdout.strip())
         res.created(rid, info)
 
@@ -183,49 +221,53 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
     def l(self, s):
         self.log.debug(s)
 
-    def setUp(self):
-        '''Create some devices in the portal to test'''
-        self.log = logging.getLogger("TestRPC")
-        self.portalcik = config['portalcik']
-        self.client = Resource(
-            self.portalcik,
-            'client',
-            {'limits': {'dataport': 'inherit',
-                        'datarule': 'inherit',
-                        'dispatch': 'inherit',
-                        'disk': 'inherit',
-                        'io': 'inherit'},
-            'writeinterval': 'inherit',
-            "name": "testclient",
-            "visibility": "parent"})
-        self._createMultiple(self.portalcik, [self.client])
-
-        # test details for create, read, and write tests.
-        cik = self.client.cik()
-        self.dataports = {}
-        self.dataports['integer'] = Resource(
+    def _createDataports(self, cik=None):
+        # test one of each type of dataport
+        if cik is None:
+            cik = self.client.cik()
+        stdports = {}
+        stdports['integer'] = Resource(
             cik, 'dataport', {'format': 'integer', 'name': 'int_port'},
             write=['-1', '0', '100000000'],
             record=[[665366400, '42']])
-        self.dataports['boolean'] = Resource(
+        stdports['boolean'] = Resource(
             cik, 'dataport', {'format': 'boolean', 'name': 'boolean_port'},
             write=['false', 'true', 'false'],
             record=[[-100, 'true'], [-200, 'false'], [-300, 'true']])
-        self.dataports['string'] = Resource(
+        stdports['string'] = Resource(
             cik, 'dataport', {'format': 'string', 'name': 'string_port'},
             alias='string_port_alias',
             write=['test', 'a' * 300],
             record=[[163299600, 'home brew'], [543212345, 'nonsense']])
-        self.dataports['float'] = Resource(
+        stdports['float'] = Resource(
             cik, 'dataport', {'format': 'float', 'name': 'float_port'},
             write=['-0.1234567', '0', '3.5', '100000000.1'],
             record=[[-100, '-0.1234567'], [-200, '0'], [-300, '3.5'], [-400, '10000000.1']])
             # TODO: handle scientific notation from OneP '-0.00001'
         # TODO: handle binary dataport
 
-        self.resources = self.dataports.values()
+        self._createMultiple(cik, stdports.values())
 
-        self._createMultiple(cik, self.resources)
+        return stdports
+
+
+    def setUp(self):
+        '''Create some devices in the portal to test'''
+        self.log = logging.getLogger('TestRPC')
+        self.portalcik = config['portalcik']
+        self.client = Resource(
+            self.portalcik,
+            'client',
+            {'limits': {'client': 'inherit',
+                        'dataport': 'inherit',
+                        'datarule': 'inherit',
+                        'dispatch': 'inherit',
+                        'disk': 'inherit',
+                        'io': 'inherit'},
+            'writeinterval': 'inherit',
+            'name': 'testclient',
+            'visibility': 'parent'})
+        self._createMultiple(self.portalcik, [self.client])
 
     def tearDown(self):
         '''Clean up any test client'''
@@ -257,7 +299,8 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
 
     def write_test(self):
         '''Write command'''
-        for res in self.resources:
+        stdports = self._createDataports()
+        for res in stdports.values():
             if res.type == 'dataport' and res.write is not None:
                 # test writing
                 if res.write is not None:
@@ -270,7 +313,6 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
                     readvalues = self._readBack(res, len(res.write))
                     self._verifyWrite(res.write, readvalues)
 
-
     def _verifyRecord(self, writetime, wrotevalues, readvalues):
         '''Checks readvalues against wrotevalues and returns True if they match
         or False if they don't. This function is complicated because wrotevalues
@@ -311,6 +353,7 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
 
     def record_test(self):
         '''Record command'''
+        stdports = self._createDataports()
         def _recordAndVerify(res, recordfn):
             if res.record is not None:
                 writetime = int(time.time())
@@ -345,7 +388,7 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
                     stdin='\n'.join(['{0},{1}'.format(t, v) for t, v in res.record]))
             self.assertTrue(r.exitcode == 0)
 
-        for r in self.resources:
+        for r in stdports.values():
             if r.type == 'dataport':
                 _recordAndVerify(r, one_by_one)
                 _flush(r)
@@ -357,20 +400,42 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
     def tree_test(self):
         '''Tree command'''
         cik = self.client.cik()
+
+        r = rpc('create', cik, '--type=client', '--cikonly')
+        self.ok(r, 'create child')
+        childcik = r.stdout
+
+        stdports = self._createDataports(childcik)
+
         r = rpc('tree', cik)
         # call did not fail
-        self.assertTrue(r.exitcode == 0)
+        self.ok(r, 'tree shouldn\'t fail')
         # starts with cik
         self.l(r.stdout)
         self.assertTrue(
             re.match("cik: {0}.*".format(cik), r.stdout) is not None)
+
         # has correct number of lines
-        self.assertTrue(len(r.stdout.split('\n')) == len(self.resources) + 1)
+        self.assertTrue(len(r.stdout.split('\n')) == len(stdports) + 1 + 1)
+
+        r = rpc('tree', cik, '--level=0')
+        self.ok(r, 'tree with --level=0 shouldn\'t fail')
+        self.ok(r)
+        self.assertTrue(len(r.stdout.split('\n')) == 1)
+
+        r = rpc('tree', cik, '--level=1')
+        self.ok(r, 'tree with --level=1 shouldn\'t fail')
+        self.assertTrue(len(r.stdout.split('\n')) == 2)
+
+        r = rpc('tree', cik, '--level=2')
+        self.ok(r, 'tree with --level=2 shouldn\'t fail')
+        self.assertTrue(len(r.stdout.split('\n')) == len(stdports) + 1 + 1)
 
     def map_test(self):
         '''Map/unmap commands'''
+        stdports = self._createDataports()
         cik = self.client.cik()
-        for res in self.resources:
+        for res in stdports.values():
             alias = 'foo'
             r = rpc('info', cik, alias)
             self.assertTrue(r.exitcode == 1, "info with alias should not work")
@@ -466,36 +531,98 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         '''Script upload'''
         waitsec = 12
         cik = self.client.cik()
-        r = rpc('script', 'files/helloworld.lua', cik)
-        self.assertTrue(r.exitcode == 0, 'New script')
-        time.sleep(waitsec)
-        self._latest(cik, 'helloworld.lua', 'line 1: Hello world!',
-                     'debug output within {0} sec'.format(waitsec))
-        self._latest(cik, 'string_port_alias', 'Hello dataport!',
-                     'dataport write from script within {0} sec'.format(waitsec))
-        r = rpc('script', 'files/helloworld2.lua', cik, '--name={0}'.format('helloworld.lua'))
-        self.assertTrue(r.exitcode == 0, 'Update existing script')
-        time.sleep(waitsec)
-        self._latest(cik, 'helloworld.lua', 'line 1: Hello world 2!',
-                     'debug output from updated script within {0} sec'.format(waitsec))
-        self._latest(cik, 'string_port_alias', 'Hello dataport 2!',
-                     'dataport write from updated script within {0} sec'.format(waitsec))
+        desc = json.dumps({'limits': {'client': 1,
+                                      'dataport': 'inherit',
+                                      'datarule': 'inherit',
+                                      'dispatch': 'inherit',
+                                      'disk': 'inherit',
+                                      'io': 'inherit'},
+            'writeinterval': 'inherit',
+            'name': 'testclient',
+            'visibility': 'parent'})
+        r = rpc('create', cik, '--type=client', '--name=firstChild', '-', stdin=desc)
+        self.ok(r, 'create child 1')
+        childrid1, childcik1 = self._ridcik(r.stdout)
+        r = rpc('create', cik, '--type=client', '--name=secondChild', '-', stdin=desc)
+        self.ok(r, 'create child 2')
+        childrid2, childcik2 = self._ridcik(r.stdout)
+        r = rpc('create', childcik2, '--type=client', '--name=grandChild')
+        self.ok(r, 'create grandchild')
+        childrid3, childcik3 = self._ridcik(r.stdout)
+
+        lua1 = {'name': 'helloworld.lua',
+                'path': 'files/helloworld.lua',
+                'out': 'line 1: Hello world!',
+                'portoutput': 'Hello dataport!'}
+        lua1['content'] = open(lua1['path']).read().strip()
+        lua2 = {'name': 'helloworld2.lua',
+                'path': 'files/helloworld2.lua',
+                'out': 'line 1: Hello world!',
+                'portoutput': 'Hello dataport 2!'}
+        lua2['content'] = open(lua2['path']).read().strip()
+
+        def readscript(cik, alias):
+            r = rpc('info', cik, alias, '--include=description')
+            self.l(r.exitcode)
+            self.l(r.stdout)
+            self.l(r.stderr)
+            info = json.loads(r.stdout)
+            return info['description']['rule']['script'].strip()
+
+        r = rpc('script', lua1['path'], cik)
+        r = rpc('read', cik, lua1['name'])
+        self.notok(r, "Don't create script unless --create passed")
+        r = rpc('script', lua1['path'], '--create', cik)
+        self.ok(r, 'New script')
+        self.assertEqual(readscript(cik, lua1['name']), lua1['content'])
+        #self._latest(cik, lua1['name'], lua1['out'],
+        #             'debug output within {0} sec'.format(waitsec))
+        #self._latest(cik, 'string_port_alias', lua1['portoutput'],
+        #             'dataport write from script within {0} sec'.format(waitsec))
+        r = rpc('script', lua2['path'], cik, '--name=' + lua1['name'])
+        self.ok(r, 'Update existing script')
+        self.assertEqual(readscript(cik, lua1['name']), lua2['content'])
+        #self._latest(cik, lua1['name'], lua2['out'],
+        #             'debug output from updated script within {0} sec'.format(waitsec))
+        #self._latest(cik, 'string_port_alias', lua2['portoutput'],
+        #             'dataport write from updated script within {0} sec'.format(waitsec))
+
+        # test --recursive
+        r = rpc('read', childcik1, lua1['name'])
+        self.notok(r, 'not recursive when --recursive is not passed')
+        r = rpc('script', lua1['path'], '--create', childcik2)
+        self.ok(r, 'create script in one child')
+        r = rpc('script', lua1['path'], '--recursive', cik)
+        self.ok(r, 'recursively write a script')
+        self.assertEqual(readscript(cik, lua1['name']), lua1['content'])
+        self.assertEqual(readscript(childcik2, lua1['name']), lua1['content'])
+        r = rpc('read', childcik1, lua1['name'])
+        self.notok(r, 'script should not be created when --create is not passed')
+        r = rpc('script', lua2['path'], cik, '--name=' + lua1['name'], '--recursive', '--create')
+        self.ok(r, 'recursive script update to helloworld2')
+        self.assertEqual(readscript(cik, lua1['name']), lua2['content'])
+        self.assertEqual(readscript(childcik1, lua1['name']), lua2['content'], "child1 updated")
+        self.assertEqual(readscript(childcik2, lua1['name']), lua2['content'], "child2 updated")
+        self.assertEqual(readscript(childcik3, lua1['name']), lua2['content'], "grandchild updated")
+
 
     def usage_test(self):
         '''OneP resource usage'''
-        r = rpc('usage', self.client.cik(), '--start=10/1/2012', '--end=11/1/2013')
-        self.assertTrue(r.exitcode == 0, 'usage call succeeded')
-        s1 = r.stdout
-        r = rpc('usage', self.client.cik(), '--start=10/1/2012', '--end=1383282000')
-        self.assertTrue(r.exitcode == 0, 'usage call succeeded')
-        s2 = r.stdout
-        r = rpc('usage', self.client.cik(), '--start=1349067600', '--end=1383282000')
-        self.assertTrue(r.exitcode == 0, 'usage call succeeded')
-        s3 = r.stdout
-        self.l(s1)
-        self.l(s2)
-        self.l(s3)
-        self.assertTrue(s1 == s2 and s2 == s3, 'various date forms output matches')
+        # This test passes inconsistently due to time passing between calls to
+        # usage. Mainly all it was testing was date parsing, though.
+        #r = rpc('usage', self.client.cik(), '--start=10/1/2012', '--end=11/1/2013')
+        #self.assertTrue(r.exitcode == 0, 'usage call succeeded')
+        #s1 = r.stdout
+        #r = rpc('usage', self.client.cik(), '--start=10/1/2012', '--end=1383282000')
+        #self.assertTrue(r.exitcode == 0, 'usage call succeeded')
+        #s2 = r.stdout
+        #r = rpc('usage', self.client.cik(), '--start=1349067600', '--end=1383282000')
+        #self.assertTrue(r.exitcode == 0, 'usage call succeeded')
+        #s3 = r.stdout
+        #self.l(s1)
+        #self.l(s2)
+        #self.l(s3)
+        #self.assertTrue(s1 == s2 and s2 == s3, 'various date forms output matches')
         def parse_metric(metric, r):
             self.assertTrue(r.exitcode == 0, 'usage call succeeded')
             self.l(r.stdout)
@@ -507,17 +634,17 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         self._create(Resource(self.client.cik(),
                               'dataport',
                               {'format': 'integer', 'name': 'int_port'}))
+        # note that this measures seconds that the dataport existed, so time
+        # must pass for the value to go up.
+        time.sleep(1)
         r = rpc('usage', self.client.cik(), '--start=10/1/2012T13:04:05', '--end=now')
         dp2 = parse_metric('dataport', r)
         self.l("dp1: {0} dp2: {1}".format(dp1, dp2))
-        # TODO: why does this not increase consistently?
-        # note that this measures seconds that the dataport existed, so time
-        # must pass for the value to go up.
-        time.sleep(5)
         self.assertTrue(dp2 > dp1, 'adding dataport added to dataport metric')
 
     def readmultiple_test(self):
         '''Read multiple RIDs'''
+        stdports = self._createDataports()
         dataports = []
         strings =  [('2013-07-20T02:40:07', 'a'),
                     ('2013-07-20T02:50:07', 'b'),
@@ -530,7 +657,7 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
                     ('2013-07-20T03:00:08', 0.3)]
         cik = self.client.cik()
         def rec(fmt, data):
-            r = rpc('record', cik, self.dataports[fmt].rid,
+            r = rpc('record', cik, stdports[fmt].rid,
                 *['--value={0},{1}'.format(t, v) for t, v in data])
             self.assertTrue(r.exitcode == 0)
 
@@ -538,7 +665,7 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         rec('integer', integers)
         rec('float', floats)
 
-        rids = [self.dataports[fmt].rid for fmt in ['string', 'integer', 'float']]
+        rids = [stdports[fmt].rid for fmt in ['string', 'integer', 'float']]
         r = rpc('read', '--start=2013-07-20T3:00:08', '--end=2013-07-20T3:00:08', cik, *rids)
         self.assertEqual(r.exitcode, 0, 'read with multiple rids')
         self.assertTrue(r.stdout == '2013-07-20 03:00:08,,3,0.3', 'two readings on one timestamp')
@@ -558,60 +685,53 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         self.ok(r, 'rid order reversed')
         self.assertTrue(r.stdout == '2013-07-20 03:00:08,0.3,3,', 'rid order reversed')
 
-
-    def _logall(self, r):
-        self.l('stdout: {0}\nstderr: {1}'.format(r.stdout, r.stderr))
-
-    def _stdre(self, r, msg, search, match, stderr=False):
-        std, label = (r.stderr, "stderr") if stderr else (r.stdout, "stdout")
-        if search is not None:
-            self.assertTrue(re.search(search, std, flags=re.MULTILINE) is not None,
-                msg + ' - failed to find {0}\n'.format(search) + label + ':' + std + '\nsearch expression: ' + search)
-        if match is not None:
-            self.assertTrue(re.match(match, std, flags=re.MULTILINE) is not None,
-                msg + ' - failed to match {0}\n'.format(match) + label + ':' + std + '\nmatch expression: ' + match)
-
-    def notok(self, response, msg='', search=None, match=None):
-        if response.exitcode == 0:
-            self._logall(response)
-        self.assertNotEqual(response.exitcode, 0, msg + ' (exit code should not be 0)')
-        self._stdre(response, msg, search=search, match=match, stderr=True)
-
-    def ok(self, response, msg='', search=None, match=None):
-        if response.exitcode != 0:
-            self._logall(response)
-        self.assertEqual(response.exitcode, 0, msg + ' (exit code should be 0)')
-        self._stdre(response, msg, search=search, match=match, stderr=False)
-
-
-    def copy_test(self):
+    def copy_diff_test(self):
         '''Copy and diff commands'''
-
+        stdports = self._createDataports()
         cik = self.client.cik()
 
         r = rpc('diff', cik, self.client.cik())
-        self.ok(r, 'diff with itself, no differences', match='')
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'diff with itself, no differences', match='')
 
         r = rpc('copy', cik, self.portalcik, '--cikonly')
         self.ok(r, 'copy test client', match=self.RE_RID)
         copycik = r.stdout
 
         r = rpc('diff', cik, copycik, '--no-children')
-        self.ok(r, '--no-children, no differences', match='')
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, '--no-children, no differences', match='')
+
 
         r = rpc('diff', copycik, self.client.cik(), '--no-children')
-        self.ok(r, 'reverse cik, still no differences', match='')
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'reverse cik, still no differences', match='')
 
         r = rpc('diff', copycik, cik)
-        self.ok(r, 'diff with children should match', match='')
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'diff with children should match', match='')
 
         newalias = 'newalias'
-        r = rpc('map', cik, self.dataports['string'].rid, newalias)
+        r = rpc('map', cik, stdports['string'].rid, newalias)
         self.ok(r, 'add an alias')
         r = rpc('diff', copycik, cik)
-        self.ok(r, 'diff notices new alias', search=r'^\+.*' + newalias)
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'diff notices new alias', search=r'^\+.*' + newalias)
         r = rpc('diff', cik, copycik)
-        self.ok(r, 'diff notices new alias (reversed)', search=r'^\-.*' + newalias)
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'diff notices new alias (reversed)', search=r'^\-.*' + newalias)
 
         r = rpc('lookup', copycik, 'string_port_alias')
         self.ok(r, 'lookup copy dataport')
@@ -619,7 +739,10 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         r = rpc('map', copycik, copyrid, newalias)
         self.ok(r, 'add same alias to copy')
         r = rpc('diff', cik, copycik)
-        self.ok(r, 'aliases match now', match='')
+        if sys.version_info < (2, 7):
+            self.notok(r, 'diff not supported with Python <2.7')
+        else:
+            self.ok(r, 'aliases match now', match='')
         '''
         r = rpc('copy', cik, cik, '--cikonly')
         self.ok(r, 'copy client into itself', match=self.RE_RID)
@@ -650,23 +773,58 @@ Asked for desc: {0}\ngot desc: {1}'''.format(res.desc, res.info['description']))
         r = rpc('--port=88', 'info', cik)
         self.notok(r, 'invalid port', match='JSON RPC Request Exception.*')
 
-        # TODO: test httptimeout
-        #rid = self._rid(
-        #    rpc('create', cik, '--type=dataport', '--format=string', '--ridonly').stdout)
+    def info_test(self):
+        '''Info command'''
+        allkeys = ['aliases', 'basic', 'counts', 'description', 'key',
+                   'shares', 'storage', 'subscribers', 'tags', 'usage']
+        cik = self.client.parentcik
+        rid = self.client.rid
 
-        #timeoutreadsize = 2000
-        #limit = '--limit={0}'.format(timeoutreadsize)
-        #timeout = '--httptimeout=120'
-        #smalltimeout = '--httptimeout=1'
+        # all keys at once
+        r = rpc('info', cik, rid)
+        self.ok(r, 'info on all keys')
+        info = json.loads(r.stdout)
+        for k in allkeys:
+            self.assertTrue(k in info.keys(), 'found key {0} when options is empty'.format(k))
 
-        #r = rpc(timeout, 'record', cik, rid, '--interval={0}'.format(60), '-', stdin='0123456789\n' * timeoutreadsize)
-        #self.ok(r, 'large record')
+        for k in allkeys:
+            # include each key
+            r = rpc('info', cik, rid, '--include={0}'.format(k))
+            self.ok(r, 'info --include={0}'.format(k))
+            info = json.loads(r.stdout)
+            self.assertTrue(info.keys() == [k], 'only requested key was returned')
+            # exclude each key
+            r = rpc('info', cik, rid, '--exclude={0}'.format(k))
+            self.ok(r, 'info --exclude={0}'.format(k))
+            info = json.loads(r.stdout)
+            keys = info.keys()
+            self.assertTrue(len(keys) == len(allkeys) - 1 and k not in keys)
 
-        #r = rpc('read', cik, limit)
-        #self.ok(r, 'large read with standard timeout', match=r"(.*\n){" + str(timeoutreadsize) + "}")
+    def read_test(self):
+        '''Read command'''
+        # record a large amount of data to a float datasource
+        cik = self.client.cik()
+        rid1, rid2 = self._createMultiple(cik, [
+            Resource(cik, 'dataport', {'format': 'float', 'name': 'float_port'}),
+            Resource(cik, 'dataport', {'format': 'string', 'name': 'string_port'})])
 
-        #r = rpc(timeout, 'read', cik, limit)
-        #self.ok(r, 'large read with large timeout')
+        numpts = 20000
+        intervalsec = 60
+        r = rpc('--httptimeout=480', 'record', cik, rid1,
+                '--interval={0}'.format(intervalsec),
+                '-', stdin='0.987654321\n' * numpts)
+        self.ok(r, "create data")
 
-        #r = rpc(smalltimeout, 'read', cik, limit)
-        #self.notok(r, 'large read with small timeout')
+        end = int(time.mktime(datetime.now().timetuple()))
+        start = int(end - (numpts * intervalsec))
+        self.l("{0},{1}".format(start, end))
+        readcmd = ['--httptimeout=1', 'read', cik, rid1,
+                   '--limit={0}'.format(numpts),
+                   '--start={0}'.format(start),
+                   '--end={0}'.format(end)]
+        r = rpc(*readcmd)
+        self.notok(r, "read a lot of data with a single big read")
+
+        readcmdchunks = readcmd + ['--chunkhours=6']
+        r = rpc(*readcmdchunks)
+        self.ok(r, "read a lot of data with multiple reads")
